(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: index_set_1d_union.ml,v 1.1 2013/06/04 17:11:41 fclement Exp $ *)

open Natural_range;;
open Index_set_1d;;

let glue_before before_mr1 x =
  let rec loop l1 before_mr1 =
    match before_mr1 with
    | [] -> [], l1
    | x :: xs ->
      if Unatural.succ x = l1
      then loop x xs
      else before_mr1, l1 in
  loop x before_mr1
;;

let glue_after after_mr1 x =
  let rec loop r1 after_mr1 =
    match after_mr1 with
    | [] -> [], r1
    | x :: xs ->
      if Unatural.pred x = r1
      then loop x xs
      else after_mr1, r1 in
  loop x after_mr1
;;

let range_extensional_union mr1 ia2 =
  let { Natural_range.left_bound = l1; right_bound = r1; } = mr1 in
  let len2 = Array.length ia2 in
  let l2 = ia2.(0)
  and r2 = Unatural.succ ia2.(len2 - 1) in
  if r1 < l2
  then Index_set_1d.is_union (is_range mr1, is_extensional ia2) else
  if r2 < l1
  then is_union (is_extensional ia2, is_range mr1) else
  let rec loop before_mr1 l1 r1 after_mr1 i =
    if i >= len2
    then
      let mr1 = Index_set_1d.is_range (natural_range l1 r1) in
      match (before_mr1, after_mr1) with
      | ([], []) -> mr1
      | ([], after_mr1) ->
          Index_set_1d.is_union (
            mr1,
            Index_set_1d.is_extensional (
              Array.of_list (List.rev after_mr1)))
      | (before_mr1, []) ->
          is_union (
            is_extensional (Array.of_list (List.rev before_mr1)),
            mr1)
      | (before_mr1, after_mr1) ->
          is_union
            (is_extensional
               (Array.of_list (List.rev before_mr1)),
             is_union
               (mr1, is_extensional (Array.of_list (List.rev after_mr1))))
    else
      let x = ia2.(i) in
      match compare x (Unatural.pred l1) with
      | -1 -> loop (x :: before_mr1) l1 r1 after_mr1 (i + 1)
      | 0 ->
        let before_mr1, l1 = glue_before before_mr1 x in
        loop before_mr1 l1 r1 after_mr1 (i + 1)
      | _ ->
        begin match compare x (Unatural.succ r1) with
        | -1 -> loop before_mr1 l1 r1 after_mr1 (i + 1)
        | 0 ->
          let after_mr1, r1 = glue_after after_mr1 x in
          loop before_mr1 l1 r1 after_mr1 (i + 1)
        | _ -> loop before_mr1 l1 r1 (x :: after_mr1) (i + 1)
        end in
  loop [] l1 r1 [] 0
;;

(*
if x < Unatural.pred l1 then
        loop (x :: before_mr1) l1 r1 after_mr1 (i + 1) else
      if x = Unatural.pred l1 then
        let before_mr1, l1 = glue_before before_mr1 x in
        loop before_mr1 l1 r1 after_mr1 (i + 1) else
      if x < Unatural.succ r1 then
        loop before_mr1 l1 r1 after_mr1 (i + 1) else
      if x = Unatural.succ r1 then
        let after_mr1, r1 = glue_after after_mr1 x in
        loop before_mr1 l1 r1 after_mr1 (i + 1) else
      loop before_mr1 l1 r1 (x :: after_mr1) (i + 1) in
  loop [] l1 r1 [] 0
;;
*)

let range_range_union nr1 nr2 =
  let { left_bound = l1; right_bound = r1; } = nr1 in
  let { left_bound = l2; right_bound = r2; } = nr2 in
  if r1 < l2 then is_union (is_range nr1, is_range nr2) else
  if r2 < l1 then is_union (is_range nr2, is_range nr1) else
  let l = min l1 l2
  and r = max r1 r2 in
  is_range (natural_range l r)
;;

let extensional_extensional_union ia1 ia2 =
  is_extensional (Array.concat [ia1; ia2])
;;

let rec union (s1, s2) =
  match s1, s2 with
  | Is_empty, _ -> s2
  | _, Is_empty -> s1
  | Is_range nr1, Is_range nr2 ->
    range_range_union nr1 nr2
  | Is_range nr1, Is_extensional na2 ->
    range_extensional_union nr1 na2
  | Is_range _, Is_union (s21, s22) ->
    union (union (s1, s21), s22)
  | Is_extensional na1, Is_range na2 -> assert false
  | Is_extensional na1, Is_extensional na2 ->
    extensional_extensional_union na1 na2
  | Is_extensional _, Is_union (s21, s22) ->
    union (union (s1, s21), s22)
  | Is_union _, Is_range _ -> assert false
  | Is_union _, Is_extensional _ -> assert false
  | Is_union (s11, s12), Is_union (_, _) ->
      union (s11, union (s12, s2))
;;

(*
let rec intersection (s1, s2) =
  match s1, s2 with
  | Is_empty, _ -> s1
  | _, Is_empty -> s2
  | Is_range nr1, Is_range nr2 ->
    range_range_intersection nr1 nr2
  | Is_range nr1, Is_extensional na2 ->
    range_extensional_intersection nr1 na2
  | Is_range _, Is_intersection (s21, s22) ->
    intersection (intersection (s1, s21), s22)
  | Is_extensional na1, Is_range na2 -> assert false
  | Is_extensional na1, Is_extensional na2 ->
    extensional_extensional_intersection na1 na2
  | Is_extensional _, Is_intersection (s21, s22) ->
    intersection (intersection (s1, s21), s22)
  | Is_intersection _, Is_range _ -> assert false
  | Is_intersection _, Is_extensional _ -> assert false
  | Is_intersection (s11, s12), Is_intersection (_, _) ->
      intersection (s11, intersection (s12, s2))
;;
*)
